#include <sys/types.h>
#include "bdm.h"
#include "bdmlib.h"
#include "bdmflash.h"

/* predefined flash types */

flash_alg_info_t amd_29f800_x16={
  addr_mask:  0xfffff,
  reg1_addr:  0x555*2,
  reg2_addr:  0x2aa*2,
  sec_size:   0x10000,
  width8:     0,
  cmd_unlock1:0xaa,	/* reg1 */
  cmd_unlock2:0x55,	/* reg2 */
  cmd_rdid:   0x90,	/* reg1 */
  cmd_prog:   0xa0,	/* reg1 */
  cmd_erase:  0x80,	/* reg1 */
  cmd_reset:  0xf0,	/* any */
  erase_all:  0x10,	/* reg1 */
  erase_sec:  0x30,	/* sector */
  fault_bit:  0x20,
  manid:      1,
  devid:      0x22d6,
  alg_name:   "amd29f800"
};

flash_alg_info_t amd_29f400_x16={
  addr_mask:  0x7ffff,
  reg1_addr:  0x555*2,
  reg2_addr:  0x2aa*2,
  sec_size:   0x10000,
  width8:     0,
  cmd_unlock1:0xaa,	/* reg1 */
  cmd_unlock2:0x55,	/* reg2 */
  cmd_rdid:   0x90,	/* reg1 */
  cmd_prog:   0xa0,	/* reg1 */
  cmd_erase:  0x80,	/* reg1 */
  cmd_reset:  0xf0,	/* any */
  erase_all:  0x10,	/* reg1 */
  erase_sec:  0x30,	/* sector */
  fault_bit:  0x20,
  manid:      1,
  devid:      0x22ab,
  alg_name:   "amd29f400"
};

/* please keep sorted by flash sizes, smallest first */
flash_alg_info_t *flash_alg_infos_def[]={
  &amd_29f400_x16,
  &amd_29f800_x16,
  NULL
};

flash_alg_info_t **flash_alg_infos=flash_alg_infos_def;

#if 0
  #define FLASH_WR16(adr,val) (*(volatile u_int16_t*)(adr)=(val))
  #define FLASH_RD16(adr) (*(volatile u_int16_t*)(adr))
#else
  #define FLASH_WR16(adr,val) \
  		({ \
  		  if(bdmlib_write_var(adr,BDM_SIZE_WORD,val)<0) \
  		  	goto mem_op_error; \
  		  val; \
  		})
  #define FLASH_RD16(adr) \
  		({ u_int16_t temp_val; \
  		  if(bdmlib_read_var(adr,BDM_SIZE_WORD,&temp_val)<0) \
  		  	goto mem_op_error; \
  		  temp_val; \
  		})
#endif

int bdmflash_check_id(flash_alg_info_t const *alg, void *addr, 
		   u_int16_t retid[2])
{
  int ret=0;
  u_int16_t devid, manid;
  caddr_t a=(caddr_t)((u_int32_t)addr&~alg->addr_mask);
  /* reset */
  FLASH_WR16(a,alg->cmd_reset);
  /* security sequence */
  FLASH_WR16(a+alg->reg1_addr,alg->cmd_unlock1);
  if(alg->cmd_unlock2)
    FLASH_WR16(a+alg->reg2_addr,alg->cmd_unlock2);
  /* read manufacturer ID */
  FLASH_WR16(a+alg->reg1_addr,alg->cmd_rdid);
  manid=FLASH_RD16(a+0);
  if(manid!=alg->manid) ret=-1;
  /* reset */
  FLASH_WR16(a,alg->cmd_reset);
  /* security sequence */
  FLASH_WR16(a+alg->reg1_addr,alg->cmd_unlock1);
  if(alg->cmd_unlock2)
    FLASH_WR16(a+alg->reg2_addr,alg->cmd_unlock2);
  /* read device ID */
  FLASH_WR16(a+alg->reg1_addr,alg->cmd_rdid);
  devid=FLASH_RD16(a+2);
  if(devid!=alg->devid) ret=-1;
  /* reset */
  FLASH_WR16(a,alg->cmd_reset);
  if(retid)
    {retid[0]=manid;retid[1]=devid;};
  return ret;

  mem_op_error:
    fprintf(stderr, "bdm memory access error\n");
    return -5;
}

int bdmflash_prog(flash_alg_info_t const *alg, void *addr, u_int16_t val)
{
  int ret=0;
  u_int16_t old,new;
  caddr_t a=(caddr_t)((u_int32_t)addr&~alg->addr_mask);
  /* security sequence */
  FLASH_WR16(a+alg->reg1_addr,alg->cmd_unlock1);
  if(alg->cmd_unlock2)
    FLASH_WR16(a+alg->reg2_addr,alg->cmd_unlock2);
  /* program command */
  FLASH_WR16(a+alg->reg1_addr,alg->cmd_prog);
  FLASH_WR16(addr,val);
  /* wait for result */
  old=FLASH_RD16(addr);
  while((new=FLASH_RD16(addr))!=old){
    if((old&alg->fault_bit)&&(new&alg->fault_bit)){
      if((FLASH_RD16(addr))!=new) ret=-2;
      break;
    }
    old=new;
  }
  /* reset */
  FLASH_WR16(a,alg->cmd_reset);
  if(FLASH_RD16(addr)!=val) ret--;
  return ret;

  mem_op_error:
    fprintf(stderr, "bdm memory access error\n");
    return -5;
}

int bdmflash_erase(flash_alg_info_t const *alg, void *addr, long size)
{
  u_int16_t old,new;
  int ret=0;
  caddr_t a=(caddr_t)((u_int32_t)addr&~alg->addr_mask);
  /* reset */
  FLASH_WR16(a,alg->cmd_reset);
  /* security sequence */
  FLASH_WR16(a+alg->reg1_addr,alg->cmd_unlock1);
  if(alg->cmd_unlock2)
    FLASH_WR16(a+alg->reg2_addr,alg->cmd_unlock2);
  /* erase command */
  FLASH_WR16(a+alg->reg1_addr,alg->cmd_erase);
  /* security sequence */
  FLASH_WR16(a+alg->reg1_addr,alg->cmd_unlock1);
  if(alg->cmd_unlock2)
    FLASH_WR16(a+alg->reg2_addr,alg->cmd_unlock2);
  /* select erase range */
  a=addr;
  if(size==0)
    FLASH_WR16(a+alg->reg1_addr,alg->erase_all);
  else{
    FLASH_WR16(addr,alg->erase_sec);
  }  
  old=FLASH_RD16(addr);
  while((new=FLASH_RD16(addr))!=old){
    if((old&alg->fault_bit)&&(new&alg->fault_bit)){
      if((FLASH_RD16(addr))!=new) ret=-2;
      break;
    }
    old=new;
  }
  /* reset */
  FLASH_WR16(a,alg->cmd_reset);
  if(FLASH_RD16(addr)!=0xffff) ret--;    
  return ret;

  mem_op_error:
    fprintf(stderr, "bdm memory access error\n");
    return -5;
}

flash_alg_info_t const *
bdmflash_alg_from_id(u_int16_t id[2])
{
  int i;
  flash_alg_info_t const *alg;
  for(i=0;(alg=flash_alg_infos_def[i])!=NULL;i++)
    if((alg->manid==id[0])&&(alg->devid==id[0]))
      return alg;
  return NULL;
}

flash_alg_info_t const *
bdmflash_alg_probe(caddr_t flash_adr)
{
  int i;
  flash_alg_info_t const *alg,*alg1;
  u_int16_t testid[2];
  for(i=0;(alg=flash_alg_infos_def[i])!=NULL;i++){
    if(bdmflash_check_id(alg,flash_adr,testid)>=0)
      return alg;
    alg1=bdmflash_alg_from_id(testid);
    if(alg1!=NULL)
      if(bdmflash_check_id(alg1,flash_adr,testid)>=0)
        return alg1;
  }
  return NULL;
}

int
bdmflash_wrb_filt(bdmlib_bfilt_t * filt, caddr_t in_adr,
		u_int size, u_char * bl_ptr)
{
  int offs=0,res;
  u_int16_t val;
  flash_alg_info_t const *alg=(flash_alg_info_t*)filt->info;

  if((u_long)in_adr&1) {
    in_adr-=1;
    val=(FLASH_RD16(in_adr)&0xff00)|bl_ptr[offs];
    if((res=bdmflash_prog(alg, in_adr, val))<0) {
      fprintf(stderr, "flash byte write error %d at 0x%lx\n",
                      res,(u_long)in_adr);
      return offs;
    }
    in_adr+=2;
    size-=1;
    offs+=1;
  }
  while(size>=2) {
    val=(bl_ptr[offs]<<8)|bl_ptr[offs+1];
    if((res=bdmflash_prog(alg, in_adr, val))<0) {
      fprintf(stderr, "flash write error %d at 0x%lx\n",
                      res,(u_long)in_adr);
      return offs;
    }
    in_adr+=2;
    size-=2;
    offs+=2;
  }
  if(size) {
    val=(FLASH_RD16(in_adr)&0x00ff)|(bl_ptr[offs]<<8);
    if((res=bdmflash_prog(alg, in_adr, val))<0) {
      fprintf(stderr, "flash byte write error %d at 0x%lx\n",
                      res,(u_long)in_adr);
      return offs;
    }
    size-=1;
    offs+=1;
  }
  return offs;

  mem_op_error:
    fprintf(stderr, "bdm memory access error\n");
    return 0;
}

int
bdmflash_erase_filt(bdmlib_bfilt_t * filt, caddr_t in_adr,
		u_int size)
{
  int res=0;
  flash_alg_info_t const *alg=(flash_alg_info_t*)filt->info;
  if(!in_adr) in_adr=filt->begin_adr;

  res=bdmflash_erase(alg, in_adr,size);
  if(res<0)
    fprintf(stderr, "flash erase error %d\n",res);
  return res;
}

